from someweta import averaged_structured_perceptron
from someweta import tagger

AveragedStructuredPerceptron = averaged_structured_perceptron.AveragedStructuredPerceptron
ASPTagger = tagger.ASPTagger



#!/usr/bin/env python3

import base64
import collections
import functools
import gzip
import json
import math
import re

import numpy as np

from someweta.averaged_structured_perceptron import AveragedStructuredPerceptron


class ASPTagger(AveragedStructuredPerceptron):
    """A part-of-speech tagger based on the averaged structured
    perceptron.

    """
    def __init__(self, beam_size, iterations, lexicon=None, mapping=None, brown_clusters=None, word_to_vec=None, ignore_tag=None):
        super().__init__(beam_size=beam_size, beam_history=2, iterations=iterations, latent_features=None, ignore_target=ignore_tag)
        self.vocabulary = set()
        self.lexicon = lexicon
        self.mapping = mapping
        if self.mapping is not None and self.ignore_target is not None:
            self.mapping[self.ignore_target] = self.ignore_target
        self.brown_clusters = brown_clusters
        self.word_to_vec = word_to_vec
        self.email = re.compile(r"^[[:alnum:].%+-]+(?:@| \[?at\]? )[[:alnum:].-]+(?:\.| \[?dot\]? )[[:alpha:]]{2,}$", re.IGNORECASE)
        self.xmltag = re.compile(r"^</?[^>]+>$")
        self.url = re.compile(r"^(?:(?:(?:https?|ftp|svn)://|(?:https?://)?www\.).+)|(?:[\w./-]+\.(?:de|com|org|net|edu|info|gov|jpg|png|gif|log|txt|xlsx?|docx?|pptx?|pdf)(?:-\w+)?)$", re.IGNORECASE)
        self.mention = re.compile(r"^@\w+$")
        self.hashtag = re.compile(r"^#\w+$")
        self.action_word = re.compile(r"^[*+][^*]+[*]$")
        self.punctuation = re.compile(r'^[](){}.!?…<>%‰€$£₤¥°@~*„“”‚‘"\'`´»«›‹,;:/*+=&%§~#^−–-]+$')
        self.ordinal = re.compile(r"^(?:\d+\.)+$")
        self.number = re.compile(r"""(?<!\w)
                                (?:[−+-]?              # optional sign
                                  \d*                  # optional digits before decimal point
                                  [.,]?                # optional decimal point
                                  \d+                  # digits
                                  (?:[eE][−+-]?\d+)?   # optional exponent
                                  |
                                  \d+[\d.,]*\d+)
                                (?![.,]?\d)""", re.VERBOSE)
        emoticon_set = set(["(-.-)", "(T_T)", "(♥_♥)", ")':", ")-:",
                            "(-:", ")=", ")o:", ")x", ":'C", ":/", ":<",
                            ":C", ":[", "=(", "=)", "=D", "=P", ">:",
                            "D':", "D:", "\:", "]:", "x(", "^^", "o.O",
                            "oO", "\O/", "\m/", ":;))", "_))", "*_*",
                            "._.", ":wink:", ">_<", "*<:-)", ":!:",
                            ":;-))"])
        emoticon_list = sorted(emoticon_set, key=len, reverse=True)
        self.emoticon = re.compile(r"""^(?:(?:[:;]|(?<!\d)8)           # a variety of eyes, alt.: [:;8]
                                    [-'oO]?                       # optional nose or tear
                                    (?: \)+ | \(+ | [*] | ([DPp])\1*(?!\w)))   # a variety of mouths
                                    """ +
                                   r"|" +
                                   r"(?:xD+|XD+)" +
                                   r"|" +
                                   r"([:;])[ ]+([()])" +
                                   r"|" +
                                   r"\^3"
                                   r"|" +
                                   r"|".join([re.escape(_) for _ in emoticon_list]) +
                                   r"$", re.VERBOSE)
        # Unicode emoticons and other symbols
        self.emoji = re.compile(r"^[\u2600-\u27BF\uFE0E\uFE0F\U0001F300-\U0001f64f\U0001F680-\U0001F6FF\U0001F900-\U0001F9FF]$")

    def train(self, words, tags, lengths):
        """"""
        lower_words = [w.lower() for w in words]
        self.latent_features = functools.partial(self._get_latent_features, lower_words)
        self.vocabulary.update(set(words))
        # self.vocabulary.update(set(lower_words))
        # <OOV>
        # # vocabulary = all lower case word forms except hapax legomena
        # self.vocabulary.update(set(k for k, v in collections.Counter(lower_words).items() if v > 1))
        # </OOV>
        X = self._get_static_features(words, lengths)
        self.fit(X, tags, lengths)

    def tag(self, words, lengths):
        """"""
        self.latent_features = functools.partial(self._get_latent_features, [w.lower() for w in words])
        X = self._get_static_features(words, lengths)
        tags = self.predict(X, lengths)
        start = 0
        for length, local_tags in zip(lengths, tags):
            local_words = words[start:start + length]
            start += length
            if self.mapping is not None:
                yield zip(local_words, local_tags, (self.mapping[lt] for lt in local_tags))
            else:
                yield zip(local_words, local_tags)

    def tag_sentence(self, sentence):
        """"""
        sentence_length = [len(sentence)]
        self.latent_features = functools.partial(self._get_latent_features, [w.lower() for w in sentence])
        X = self._get_static_features(sentence, sentence_length)
        tags = list(self.predict(X, sentence_length))[0]
        if self.mapping is not None:
            return list(zip(sentence, tags, (self.mapping[lt] for lt in tags)))
        else:
            return list(zip(sentence, tags))

    def evaluate(self, words, tags, lengths):
        """"""
        self.latent_features = functools.partial(self._get_latent_features, [w.lower() for w in words])
        X = self._get_static_features(words, lengths)
        # accuracy = self.score(X, tags, lengths)
        # return accuracy
        predicted = self.predict(X, lengths)
        correct, correct_iv, correct_oov = 0, 0, 0
        coarse_correct, coarse_correct_iv, coarse_correct_oov = 0, 0, 0
        total, total_iv, total_oov = 0, 0, 0
        start = 0
        for length, local_pred in zip(lengths, predicted):
            local_words = words[start:start + length]
            local_gold = tags[start:start + length]
            start += length
            for w, g, p in zip(local_words, local_gold, local_pred):
                if self.ignore_target is not None and g == self.ignore_target:
                    continue
                total += 1
                # <OOV>
                # if w.lower() in self.vocabulary:
                # </OOV>
                if w in self.vocabulary:
                    total_iv += 1
                    if g == p:
                        correct += 1
                        correct_iv += 1
                    if self.mapping is not None:
                        if self.mapping[g] == self.mapping[p]:
                            coarse_correct += 1
                            coarse_correct_iv += 1
                else:
                    total_oov += 1
                    if g == p:
                        correct += 1
                        correct_oov += 1
                    if self.mapping is not None:
                        if self.mapping[g] == self.mapping[p]:
                            coarse_correct += 1
                            coarse_correct_oov += 1

        accuracy = correct / total
        try:
            accuracy_iv = correct_iv / total_iv
        except ZeroDivisionError:
            accuracy_iv = 0
        try:
            accuracy_oov = correct_oov / total_oov
        except ZeroDivisionError:
            accuracy_oov = 0
        coarse_accuracy, coarse_accuracy_iv, coarse_accuracy_oov = None, None, None
        if self.mapping is not None:
            coarse_accuracy = coarse_correct / total
            try:
                coarse_accuracy_iv = coarse_correct_iv / total_iv
            except ZeroDivisionError:
                coarse_accuracy_iv = 0
            try:
                coarse_accuracy_oov = coarse_correct_oov / total_oov
            except ZeroDivisionError:
                coarse_accuracy_oov = 0
        return accuracy, accuracy_iv, accuracy_oov, coarse_accuracy, coarse_accuracy_iv, coarse_accuracy_oov

    def save(self, filename):
        """"""
        with gzip.open(filename, 'wb') as f:
            features = sorted(self.weights.keys())
            f.write("[\n".encode())
            f.write(json.dumps(list(self.vocabulary), ensure_ascii=False, indent=4).encode())
            f.write(",\n".encode())
            f.write(json.dumps(self.lexicon, ensure_ascii=False, indent=4).encode())
            f.write(",\n".encode())
            f.write(json.dumps(self.brown_clusters, ensure_ascii=False, indent=4).encode())
            f.write(",\n".encode())
            f.write(json.dumps(self.word_to_vec, ensure_ascii=False, indent=4).encode())
            f.write(",\n".encode())
            f.write(json.dumps(self.target_mapping, ensure_ascii=False, indent=4).encode())
            f.write(",\n".encode())
            f.write(str(self.target_size).encode())
            f.write(",\n".encode())
            f.write(json.dumps(features, ensure_ascii=False, indent=4).encode())
            f.write(",\n".encode())
            f.write("[\n".encode())
            for feat in features[:-1]:
                f.write('"'.encode())
                f.write(base64.b85encode(self.weights[feat].tostring()))
                f.write('",\n'.encode())
            f.write('"'.encode())
            f.write(base64.b85encode(self.weights[features[-1]].tostring()))
            f.write('"\n]\n'.encode())
            f.write("]\n".encode())

    def load(self, filename):
        """"""
        with gzip.open(filename, 'rb') as f:
            model = json.loads(f.read().decode())
            vocabulary, self.lexicon, self.brown_clusters, self.word_to_vec, self.target_mapping, self.target_size, features, weights = model
            self.vocabulary = set(vocabulary)
            self.weights = {f: np.fromstring(base64.b85decode(w), np.float64) for f, w in zip(features, weights)}

    def load_prior_model(self, prior):
        """"""
        with gzip.open(prior, 'rb') as f:
            model = json.loads(f.read().decode())
            self.vocabulary = set(model[0])
            self.target_mapping = model[4]
            self.target_size = model[5]
            features = model[6]
            weights = model[7]
            self.prior_weights = {f: np.fromstring(base64.b85decode(w), np.float64) for f, w in zip(features, weights)}

    def _get_static_features(self, words, lengths):
        """"""
        lexicon = self.lexicon
        brown_clusters = self.brown_clusters
        word_to_vec = self.word_to_vec
        # <OOV>
        # vocabulary = self.vocabulary
        # </OOV>
        features = []
        start = 0
        for length in lengths:
            sentence = words[start:start + length]
            start += length
            tokens = ["<START-2>", "<START-1>"] + [w.lower() for w in sentence] + ["<END+1>", "<END+2>"]
            for i, word in enumerate(sentence):
                j = i + 2
                local_features = []
                w = tokens[j]
                p1 = tokens[j - 1]
                p2 = tokens[j - 2]
                n1 = tokens[j + 1]
                n2 = tokens[j + 2]
                # constant bias feature acts like a prior
                local_features.append("bias")
                # rounded logarithm of word length
                local_features.append("W_loglength: %d" % round(math.log(len(word))))
                # current word
                local_features.append("W_word: %s" % w)
                # <OOV>
                # if w in vocabulary:
                #     local_features.append("W_word: %s" % w)
                # else:
                #     local_features.append("W_word: OOV")
                # </OOV>
                # next words
                local_features.append("N1_word: %s" % n1)
                # <OOV>
                # if n1 in vocabulary:
                #     local_features.append("N1_word: %s" % n1)
                # else:
                #     local_features.append("N1_word: OOV")
                # </OOV>
                local_features.append("N2_word: %s" % n2)
                # <OOV>
                # if n2 in vocabulary:
                #     local_features.append("N2_word: %s" % n2)
                # else:
                #     local_features.append("N2_word: OOV")
                # </OOV>
                # affixes
                local_features.append("W_prefix: %s" % w[:3])
                local_features.append("W_suffix: %s" % w[-3:])
                if i >= 1:
                    local_features.append("P1_suffix: %s" % p1[-3:])
                if length - i > 1:
                    local_features.append("N1_suffix: %s" % n1[-3:])
                # word shape
                local_features.append("W_shape: %s" % self._word_shape(word))
                # Flags
                if i >= 2:
                    local_features.extend(self._word_flags(p2, "P2"))
                if i >= 1:
                    local_features.extend(self._word_flags(p1, "P1"))
                local_features.extend(self._word_flags(w, "W"))
                if length - i > 1:
                    local_features.extend(self._word_flags(n1, "N1"))
                if length - i > 2:
                    local_features.extend(self._word_flags(n2, "N2"))
                # Brown clusters
                if brown_clusters is not None:
                    # P2, P1, W, N1, N2
                    if i >= 2:
                        bc, freq = brown_clusters.get(p2, ("N/A", 0))
                        local_features.append("P2_brown: %s" % bc)
                    if i >= 1:
                        bc, freq = brown_clusters.get(p1, ("N/A", 0))
                        local_features.append("P1_brown: %s" % bc)
                    bc, freq = brown_clusters.get(w, ("N/A", 0))
                    local_features.append("W_brown: %s" % bc)
                    local_features.append("W_logfreq: %d" % freq)
                    if length - i > 1:
                        bc, freq = brown_clusters.get(n1, ("N/A", 0))
                        local_features.append("N1_brown: %s" % bc)
                    if length - i > 2:
                        bc, freq = brown_clusters.get(n2, ("N/A", 0))
                        local_features.append("N2_brown: %s" % bc)
                if word_to_vec is not None:
                    # if w in word_to_vec:
                    #     for i, d in enumerate(word_to_vec[w]):
                    #         local_features.append("W_w2v_%d: %d" % (i, round(float(d))))
                    if w in word_to_vec:
                        local_features.append("W_w2v: %s" % word_to_vec[w])
                if lexicon is not None:
                    if w in lexicon:
                        for feat in lexicon[w]:
                            local_features.append("W_lex: %s" % feat)
                    else:
                        local_features.append("W_lex: N/A")
                features.append(local_features)
        return features

    def _get_latent_features(self, words, start, beam, i):
        """"""
        # <OOV>
        # vocabulary = self.vocabulary
        # </OOV>
        features = []
        global_i = start + i
        tags = ["<START-2>", "<START-1>"] + beam
        j = i + 2
        if i >= 1:
            features.append("P1_word, P1_pos: %s, %s" % (words[global_i - 1], tags[j - 1]))
            # <OOV>
            # if words[global_i - 1] in vocabulary:
            #     features.append("P1_word, P1_pos: %s, %s" % (words[global_i - 1], tags[j - 1]))
            # else:
            #     features.append("P1_word, P1_pos: OOV, %s" % tags[j - 1])
            # </OOV>
        if i >= 2:
            features.append("P2_word, P2_pos: %s, %s" % (words[global_i - 2], tags[j - 2]))
            # <OOV>
            # if words[global_i - 2] in vocabulary:
            #     features.append("P2_word, P2_pos: %s, %s" % (words[global_i - 2], tags[j - 2]))
            # else:
            #     features.append("P2_word, P2_pos: OOV, %s" % tags[j - 2])
            # </OOV>
        features.append("P1_pos: %s" % tags[j - 1])
        features.append("P2_pos: %s" % tags[j - 2])
        features.append("P2_pos, P1_pos: %s, %s" % (tags[j - 2], tags[j - 1]))
        features.append("P1_pos, W_word: %s, %s" % (tags[j - 1], words[global_i]))
        # <OOV>
        # if words[global_i] in vocabulary:
        #     features.append("P1_pos, W_word: %s, %s" % (tags[j - 1], words[global_i]))
        # else:
        #     features.append("P1_pos, W_word: %s, OOV" % tags[j - 1])
        # </OOV>
        return features

    @staticmethod
    @functools.lru_cache(maxsize=10240)
    def _word_shape(word):
        if len(word) >= 100:
            return "LONG"
        shape = []
        last = ""
        shape_char = ""
        seq = 0
        for c in word:
            if c.isalpha():
                if c.isupper():
                    shape_char = "X"
                else:
                    shape_char = "x"
            elif c.isdigit():
                shape_char = "d"
            else:
                shape_char = c
            if shape_char == last:
                seq += 1
            else:
                seq = 0
                last = shape_char
            if seq < 4:
                shape.append(shape_char)
        return "".join(shape)

    @functools.lru_cache(maxsize=10240)
    def _word_flags(self, word, prefix):
        """"""
        flags = []
        if word.isalpha():
            flags.append("%s_isalpha" % prefix)
        if word.isnumeric():
            flags.append("%s_isnumeric" % prefix)
        if word.islower():
            flags.append("%s_islower" % prefix)
        if word.isupper():
            flags.append("%s_isupper" % prefix)
        if word.istitle():
            flags.append("%s_istitle" % prefix)
        if self.email.search(word):
            flags.append("%s_isemail" % prefix)
        if self.xmltag.search(word):
            flags.append("%s_istag" % prefix)
        if self.url.search(word):
            flags.append("%s_isurl" % prefix)
        if self.mention.search(word):
            flags.append("%s_ismention" % prefix)
        if self.hashtag.search(word):
            flags.append("%s_ishashtag" % prefix)
        if self.action_word.search(word):
            flags.append("%s_isactword" % prefix)
        if self.emoticon.search(word):
            flags.append("%s_isemoticon" % prefix)
        if self.emoji.search(word):
            flags.append("%s_isemoji" % prefix)
        if self.punctuation.search(word):
            flags.append("%s_ispunct" % prefix)
        if self.ordinal.search(word):
            flags.append("%s_isordinal" % prefix)
        if self.number.search(word):
            flags.append("%s_isnumber" % prefix)
        return flags



#!/usr/bin/env python3

import collections
import itertools
import logging
import operator
import random

import numpy as np

from someweta import utils

Beam = collections.namedtuple("Beam", ["tags", "weight_sum", "features", "previous"])


class AveragedStructuredPerceptron:
    """An averaged structured perceptron.

    The perceptron algorithm is due to Rosenblatt (1958). Freund and
    Schapire (1999: 292) found "that voting and averaging perform
    better than using the last vector". We implement averaging.
    Collins (2002) introduced the structured perceptron and Collins
    and Roark (2004) suggested the early update strategy.

    """
    def __init__(self, beam_size, beam_history, iterations, latent_features, prior_weights=None, ignore_target=None):
        self.beam_size = beam_size
        self.beam_history = beam_history
        self.iterations = iterations
        self.latent_features = latent_features
        self.prior_weights = prior_weights
        self.ignore_target = ignore_target
        # self.weights = collections.defaultdict(lambda: collections.defaultdict(float))
        # self.weights_c = collections.defaultdict(lambda: collections.defaultdict(float))
        self.target_mapping = {}
        self.reverse_mapping = None
        self.target_size = 0
        self.ignore_target_mapping = None
        self.weights = {}
        self.weights_c = {}

    def fit(self, X, y, lengths):
        """"""
        targets = collections.Counter(y)
        former_target_size = self.target_size
        for target, freq in reversed(targets.most_common()):
            if target not in self.target_mapping and target != self.ignore_target:
                self.target_mapping[target] = self.target_size
                self.target_size += 1
        if self.ignore_target is not None:
            self.ignore_target_mapping = self.target_size
        y = [self.target_mapping.get(target, self.ignore_target_mapping) for target in y]
        if self.target_size > former_target_size and self.prior_weights is not None:
            for feat in self.prior_weights:
                self.prior_weights[feat].resize((self.target_size,))
        counter = 0
        ranges = list(zip((a - b for a, b in zip(itertools.accumulate(lengths), lengths)), lengths))
        for it in range(self.iterations):
            total, incorrect, early_update = 0, 0, 0
            for start, length in ranges:
                local_X = X[start:start + length]
                local_y = y[start:start + length]
                predicted, features = self._beam_search(local_X, start, local_y)
                assert type(predicted) == type(y)
                if len(predicted) != len(local_y):
                    early_update += 1
                if self.ignore_target is not None:
                    erroneous = any(p != g and g != self.ignore_target_mapping for p, g in zip(predicted, local_y))
                else:
                    erroneous = predicted != local_y
                if erroneous:
                    incorrect += 1
                    self._update(local_y, predicted, features, counter)
                counter += len(predicted)
                total += 1
            # random.seed(it)
            random.shuffle(ranges)
            correct = total - incorrect
            logging.info("Iteration %d: %d/%d = %.2f%% (%d early update)" % (it, correct, total, (correct / total) * 100, early_update))
        for feat in self.weights:
            self.weights[feat] -= self.weights_c[feat] / counter
        if self.prior_weights is not None:
            for feat in self.prior_weights:
                if feat not in self.weights:
                    self.weights[feat] = self.prior_weights[feat]
                else:
                    self.weights[feat] += self.prior_weights[feat]

    def predict(self, X, lengths):
        """"""
        if self.reverse_mapping is None:
            self.reverse_mapping = {v: k for k, v in self.target_mapping.items()}
        ranges = list(zip((a - b for a, b in zip(itertools.accumulate(lengths), lengths)), lengths))
        for start, length in ranges:
            local_X = X[start:start + length]
            predicted, features = self._beam_search(local_X, start)
            predicted = [self.reverse_mapping[p] for p in predicted]
            yield predicted

    def score(self, X, y, lengths):
        """"""
        if self.reverse_mapping is None:
            self.reverse_mapping = {v: k for k, v in self.target_mapping.items()}
        predicted = []
        ranges = list(zip((a - b for a, b in zip(itertools.accumulate(lengths), lengths)), lengths))
        for start, length in ranges:
            local_X = X[start:start + length]
            local_pred, features = self._beam_search(local_X, start)
            local_pred = [self.reverse_mapping[p] for p in local_pred]
            predicted.extend(local_pred)
        accuracy = utils.evaluate(y, predicted, self.ignore_target)
        coarse_accuracy = None
        if self.mapping is not None:
            coarse_y = [self.mapping[t] for t in y]
            coarse_predicted = [self.mapping[t] for t in predicted]
            coarse_accuracy = utils.evaluate(coarse_y, coarse_predicted, self.ignore_target)
        return accuracy, coarse_accuracy

    @staticmethod
    def _extract_feature_sequence(beam):
        """"""
        sequence = [beam.features]
        while beam.previous is not None:
            sequence.append(beam.features)
            beam = beam.previous
        return sequence[::-1]

    def _beam_search(self, X, start, y=None):
        """"""
        beams = [Beam([], 0, [], None)]
        gold_tags = []
        for i, static_features in enumerate(X):
            agenda = {}
            weight_sum = self._predict_static(static_features)
            for beam in beams:
                latent_features = self.latent_features(start, beam.tags, i)
                features = static_features + latent_features
                for prediction, weight in self._predict_latent(latent_features, weight_sum):
                    tags = beam.tags + [prediction]
                    history = tuple(tags[-self.beam_history:])
                    new_weight_sum = beam.weight_sum + weight
                    in_agenda = agenda.get(history)
                    if in_agenda is None or new_weight_sum > in_agenda.weight_sum:
                        agenda[history] = Beam(tags, new_weight_sum, features, beam)
            beams = sorted(agenda.values(), key=operator.attrgetter("weight_sum"), reverse=True)[:self.beam_size]
            if y is not None:
                gold_tags.append(y[i])
                if self.ignore_target is not None:
                    gold_not_in_beam = all(any(p != g and g != self.ignore_target_mapping for p, g in zip(beam.tags, gold_tags)) for beam in beams)
                else:
                    gold_not_in_beam = all(beam.tags != gold_tags for beam in beams)
                if gold_not_in_beam:
                    break
        return beams[0].tags, self._extract_feature_sequence(beams[0])

    def _predict_static(self, features):
        """"""
        weight_sum = np.zeros(self.target_size)
        weight_sum += np.sum([self.weights[feat] for feat in features if feat in self.weights], axis=0)
        if self.prior_weights is not None:
            weight_sum += np.sum([self.prior_weights[feat] for feat in features if feat in self.prior_weights], axis=0)
        return weight_sum

    def _predict_latent(self, features, static_weights):
        """"""
        weight_sum = np.sum([self.weights[feat] for feat in features if feat in self.weights], axis=0)
        if self.prior_weights is not None:
            weight_sum += np.sum([self.prior_weights[feat] for feat in features if feat in self.prior_weights], axis=0)
        weight_sum += static_weights
        predictions = np.argsort(weight_sum)[-self.beam_size:]
        return reversed(list(zip(predictions, weight_sum[predictions])))

    def _update(self, y, predicted, features, counter):
        """"""
        for feature_set, true_cls, predicted_cls in zip(features, y, predicted):
            if true_cls != predicted_cls:
                if self.ignore_target is not None and true_cls == self.ignore_target_mapping:
                    continue
                for feat in feature_set:
                    if feat not in self.weights:
                        self.weights[feat] = np.zeros(self.target_size)
                        self.weights_c[feat] = np.zeros(self.target_size)
                    self.weights[feat][true_cls] += 1
                    self.weights_c[feat][true_cls] += counter
                    self.weights[feat][predicted_cls] -= 1
                    self.weights_c[feat][predicted_cls] -= counter
            counter += 1

# def train_by_iterative_parameter_mixing(training_data, iterations=10, beam_size=5, n_shards=5):
#     """Iterative parameter mixing was first described by McDonald et al.
#     (2010).

#     """
#     weights = collections.defaultdict(lambda: collections.defaultdict(float))
#     weights_c = collections.defaultdict(lambda: collections.defaultdict(float))
#     counter = 0
#     random.seed(42)
#     random.shuffle(training_data)
#     div, mod = divmod(len(training_data), n_shards)
#     shards = [training_data[i * div + min(i, mod):(i + 1) * div + min(i + 1, mod)] for i in range(n_shards)]
#     for it in range(iterations):
#         ipm = functools.partial(ipm_iteration, it=it, beam_size=beam_size, w=weights, wc=weights_c, c=counter)
#         results = map(ipm, shards)
#         weights = collections.defaultdict(lambda: collections.defaultdict(float))
#         weights_c = collections.defaultdict(lambda: collections.defaultdict(float))
#         counter, correct, total, early_update = 0, 0, 0, 0
#         for w, wc, c, cor, tot, earl_upd in results:
#             for feat in w:
#                 for cls, weight in w[feat].items():
#                     weights[feat][cls] += weight / n_shards
#             for feat in wc:
#                 for cls, weight in wc[feat].items():
#                         weights_c[feat][cls] += weight / n_shards
#             counter += c / n_shards
#             correct += cor
#             total += tot
#             early_update += earl_upd
#         logging.info("Iteration %d: %d/%d = %.2f%% (%d early update)" % (it, correct, total, (correct / total) * 100, early_update))
#     # return weights - weights_c / counter
#     for feat in weights:
#         for cls in weights[feat]:
#             weights[feat][cls] -= weights_c[feat][cls] / counter
#     return weights


# def ipm_iteration(training_data, it, beam_size, w, wc, c):
#     """"""
#     weights = collections.defaultdict(lambda: collections.defaultdict(float))
#     weights_c = collections.defaultdict(lambda: collections.defaultdict(float))
#     counter = c
#     random.seed(it)
#     random.shuffle(training_data)
#     for feat in w:
#         for cls, weight in w[feat].items():
#             weights[feat][cls] = weight
#     for feat in wc:
#         for cls, weight in wc[feat].items():
#             weights_c[feat][cls] = weight
#     total, incorrect, early_update = 0, 0, 0
#     for structure in training_data:
#         items, static_features, gold_classes = structure
#         predicted, features = beam_search(items, static_features, weights, beam_size, gold_classes)
#         assert type(predicted) == type(gold_classes)
#         if len(predicted) != len(gold_classes):
#             early_update += 1
#         if predicted != gold_classes:
#             incorrect += 1
#             weights, weights_c = update(weights, weights_c, gold_classes, predicted, features, counter)
#         counter += len(predicted)
#         total += 1
#     correct = total - incorrect
#     return weights, weights_c, counter, correct, total, early_update



# 1. Upload to PyPI:
# python3 setup.py sdist
# python3 setup.py sdist upload
#
# 2. Check if everything looks all right: https://pypi.python.org/pypi/SoMeWeTa
#
# 3. Go to https://github.com/tsproisl/SoMeWeTa/releases/new and
# create a new release

from os import path
from setuptools import setup

here = path.abspath(path.dirname(__file__))
with open(path.join(here, 'README.rst')) as fh:
    long_description = fh.read()

version = "1.4.0"

setup(
    name='SoMeWeTa',
    version=version,
    author='Thomas Proisl',
    author_email='thomas.proisl@fau.de',
    packages=[
        'someweta',
    ],
    scripts=[
        'bin/somewe-tagger',
    ],
    url="https://github.com/tsproisl/SoMeWeTa",
    download_url='https://github.com/tsproisl/SoMeWeTa/archive/v%s.tar.gz' % version,
    license='GNU General Public License v3 or later (GPLv3+)',
    description='A part-of-speech tagger with support for domain adaptation and external resources.',
    long_description=long_description,
    install_requires=[
        "numpy",
    ],
    python_requires='>=3.4',
    classifiers=[
        # 'Development Status :: 3 - Alpha',
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'Intended Audience :: Science/Research',
        'License :: OSI Approved :: GNU General Public License v3 or later (GPLv3+)',
        'Natural Language :: English',
        'Natural Language :: French',
        'Natural Language :: German',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Topic :: Text Processing :: Linguistic',
    ],
)



#!/usr/bin/env python3

import collections
import html
import json
import math
import xml.etree.ElementTree as ET


def read_lexicon(filename):
    """Read in the lexicon."""
    lexicon = collections.defaultdict(set)
    with open(filename) as fh:
        for line in fh:
            line = line.strip()
            if line == "":
                continue
            word, wc = line.split("\t")
            lexicon[word.lower()].add(wc)
    for k, v in lexicon.items():
        lexicon[k] = list(v)
    return lexicon


def read_brown_clusters(fh):
    """"""
    brown_clusters = {}
    for line in fh:
        cluster, word, freq = line.rstrip().split("\t")
        brown_clusters[word] = (cluster, round(math.log(int(freq))))
    return brown_clusters


def read_mapping(filename):
    """"""
    with open(filename) as fh:
        mapping = json.load(fh)
    return mapping


def read_word2vec_vectors(fh):
    """"""
    word_to_vec = {}
    for line in fh:
        fields = line.rstrip().split()
        word_to_vec[fields[0]] = fields[1:]
    return word_to_vec


def read_word2vec_clusters(fh):
    """"""
    word_to_vec = {}
    for line in fh:
        word, cluster = line.rstrip().split("\t")
        word_to_vec[word] = cluster
    return word_to_vec


def get_sentences(fh, tagged=True):
    """A generator over the sentence in `filename`."""
    sentence = []
    for line in fh:
        line = line.strip()
        if line == "":
            if tagged:
                words, tags = zip(*sentence)
                yield list(words), list(tags)
            else:
                yield sentence
            sentence = []
        else:
            if tagged:
                sentence.append(line.split("\t", 2))
            else:
                sentence.append(line)
    if len(sentence) > 0:
        if tagged:
            words, tags = zip(*sentence)
            yield list(words), list(tags)
        else:
            yield sentence


def read_corpus(fh, tagged=True):
    """Return a list of sentences, each consisting of a list of tokens."""
    words, tags, lengths = [], [], []
    if tagged:
        for w, t, l in iter_corpus(fh, tagged):
            words.extend(w)
            tags.extend(t)
            lengths.append(l)
        return words, tags, lengths
    else:
        for w, l in iter_corpus(fh, tagged):
            words.extend(w)
            lengths.append(l)
        return words, lengths


def iter_corpus(fh, tagged=True):
    """Yield one sentence at a time, each consisting of a list of
    tokens.

    """
    for sentence in get_sentences(fh, tagged):
        if tagged:
            words, tags = sentence
            length = len(words)
            yield words, tags, length
        else:
            length = len(sentence)
            words = sentence
            yield words, length


def evaluate(gold, predicted, ignore_tag=None):
    """Evaluate accuracy of predicted against gold."""
    total = len(gold)
    correct = sum(1 for gt, pt in zip(gold, predicted) if gt == pt)
    if ignore_tag is not None:
        total = sum(1 for gt in gold if gt != ignore_tag)
    return correct / total


def parse_xml(xml, is_file=True):
    """Return a list of XML elements and their text/tail as well as the
    whole text of the document.

    """
    Element = collections.namedtuple("Element", ["element", "type", "text"])

    def text_getter(elem):
        text = elem.text
        tail = elem.tail
        if text is None:
            text = ""
        if tail is None:
            tail = ""
        yield Element(elem, "text", text)
        for child in elem:
            for t in text_getter(child):
                yield t
        yield Element(elem, "tail", tail)
    if is_file:
        tree = ET.parse(xml)
        root = tree.getroot()
    else:
        root = ET.fromstring(xml)
    elements = list(text_getter(root))
    return elements


def read_tagged_xml(xml):
    """Return a list of sentences, each consisting of a list of tokens. If
    tagged=False, also return the original lines and the indexes of
    the words.

    """
    words, tags, lengths = [], [], []
    for w, t, l in iter_xml(xml, tagged=True):
        words.extend(w)
        tags.extend(t)
        lengths.append(l)
    return words, tags, lengths


def iter_xml(xml, tagged=True):
    """Yield one sentence at a time. If tagged=False, also return the
    original lines and the indexes of the words.

    """
    for sentence in get_sentences(xml, tagged=False):
        word_indexes = [i for i, line in enumerate(sentence) if not (line.startswith("<") and line.endswith(">"))]
        if tagged:
            words, tags = zip(*[sentence[i].split("\t", 2) for i in word_indexes])
        else:
            words = [sentence[i] for i in word_indexes]
        words = [html.unescape(w) for w in words]
        length = len(words)
        if tagged:
            yield words, tags, length
        else:
            yield words, length, sentence, word_indexes


def add_pos_to_xml(tagged_sentence, lines, word_indexes):
    """Add part-of-speech tags to original lines of XML file."""
    words, tags = zip(*tagged_sentence)
    for idx, tag in zip(word_indexes, tags):
        lines[idx] += "\t%s" % tag
    return lines
